/**********************************************************************
Copyright (c) 2009 Stefan Seelmann and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors :
 ...
 ***********************************************************************/

package org.datanucleus.store.ldap.fieldmanager;

import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.naming.directory.Attributes;
import javax.naming.directory.BasicAttribute;
import javax.naming.directory.BasicAttributes;
import javax.naming.ldap.LdapName;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.EmbeddedMetaData;
import org.datanucleus.metadata.Relation;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.StoreManager;
import org.datanucleus.store.Type;
import org.datanucleus.store.ldap.LDAPUtils;

/**
 * Mapping strategy for embedded objects.
 */
public class EmbeddedMappingStrategy extends AbstractMappingStrategy
{

    protected AbstractClassMetaData effectiveClassMetaData;

    protected int fieldNumber;

    protected ExecutionContext om;
    
    protected StoreManager storeMgr;

    protected ClassLoaderResolver clr;

    protected EmbeddedMappingStrategy(StoreManager storeMgr, ObjectProvider sm, AbstractMemberMetaData mmd, Attributes attributes)
    {
        super(sm, mmd, attributes);
        this.fieldNumber = mmd.getAbsoluteFieldNumber();
        this.om = sm.getExecutionContext();
        this.storeMgr = storeMgr;
        this.clr = om.getClassLoaderResolver();
        this.effectiveClassMetaData = LDAPUtils.getEffectiveClassMetaData(mmd, sm.getExecutionContext().getMetaDataManager());

    }

    @Override
    public Object fetch()
    {
        switch (mmd.getRelationType(clr))
        {
            case Relation.ONE_TO_ONE_UNI :
            case Relation.ONE_TO_ONE_BI :

                Object value = null;
                Set<String> objectClasses = LDAPUtils.getObjectClassesForClass(effectiveClassMetaData);
                if (objectClasses.isEmpty())
                {
                    // embedded into the current entry
                    value = fetchEmbedded();
                }
                else
                {
                    // embedded as child-entry
                    value = fetchFromChild();
                }
                return value;

            case Relation.ONE_TO_MANY_UNI :
            case Relation.ONE_TO_MANY_BI :

                if (mmd.hasCollection())
                {
                    Class instanceType = mmd.getType();
                    instanceType = org.datanucleus.store.types.sco.SCOUtils.getContainerInstanceType(instanceType, mmd.getOrderMetaData() != null);
                    Collection<Object> coll = fetchFromChildren(instanceType);
                    return sm.wrapSCOField(fieldNumber, coll, false, false, true);
                }
        }

        // TODO Localise this
        throw new NucleusException(
                "Cant obtain value for field " + mmd.getFullFieldName() + " since type=" + mmd.getTypeName() + " is not supported for this datastore");

    }

    /**
     * Gets the object that is embedded within the current entry.
     * @return the object
     */
    private Object fetchEmbedded()
    {
        // use embedded meta data from field
        EmbeddedMetaData embeddedMetaData = mmd.getEmbeddedMetaData();

        // use field meta data from embedded meta data
        List<AbstractMemberMetaData> embeddedMmds = new ArrayList<AbstractMemberMetaData>(Arrays.asList(embeddedMetaData
                .getMemberMetaData()));

        ObjectProvider embeddedSM = om.newObjectProviderForMember(mmd,effectiveClassMetaData);
        embeddedSM = getEmbeddedStateManager(embeddedSM.getObject());
        Object value = fetchMerge(embeddedSM, attributes, embeddedMmds, embeddedMetaData);
        return value;
    }

    /**
     * Gets the object that is embedded from child entry under the current entry
     * @return the object
     */
    private Object fetchFromChild()
    {
        // use embedded meta data from field
        EmbeddedMetaData embeddedMetaData = mmd.getEmbeddedMetaData();
        Collection<Object> children = fetchFromChildren(ArrayList.class, mmd, embeddedMetaData);
        if (children.size() == 0)
        {
            return null;
        }
        else if (children.size() == 1)
        {
            return children.iterator().next();
        }
        else
        {
            throw new NucleusDataStoreException("Must be unique!");
        }
    }

    /**
     * Gets the objects that are embedded from child entries under the current entry.
     * @param collectionType the collection type
     * @return the objects
     */
    private Collection<Object> fetchFromChildren(Class collectionType)
    {
        // use embedded meta data from element
        EmbeddedMetaData embeddedMetaData = mmd.getElementMetaData().getEmbeddedMetaData();
        return fetchFromChildren(collectionType, mmd, embeddedMetaData);
    }

    private Collection<Object> fetchFromChildren(Class collectionType, AbstractMemberMetaData mmd, EmbeddedMetaData embeddedMetaData)
    {
        // use field meta data from class definition
        List<AbstractMemberMetaData> embeddedMmds = LDAPUtils.getAllMemberMetaData(effectiveClassMetaData);

        // search
        Collection<Object> coll = getCollectionInstance(collectionType);
        LdapName baseDn = LDAPUtils.getDistinguishedNameForObject(storeMgr, sm);
        
        // search exact type
        Map<LdapName, Attributes> entries = LDAPUtils.getEntries(storeMgr, sm.getExecutionContext(), effectiveClassMetaData, baseDn, null, false, false);
        for (Attributes embeddedAttrs : entries.values())
        {
            ObjectProvider embeddedSM = om.newObjectProviderForMember(mmd, effectiveClassMetaData);
            embeddedSM = getEmbeddedStateManager(embeddedSM.getObject());
            Object value = fetchMerge(embeddedSM, embeddedAttrs, embeddedMmds, embeddedMetaData);
            if (value != null)
            {
                coll.add(value);
            }
        }

        // search sub types
        String[] subclassNames = om.getMetaDataManager().getSubclassesForClass(effectiveClassMetaData.getFullClassName(), true);
        if (subclassNames != null)
        {
            for (String subclassName : subclassNames)
            {
                AbstractClassMetaData cmd = om.getMetaDataManager().getMetaDataForClass(subclassName, clr);
                entries = LDAPUtils.getEntries(storeMgr, sm.getExecutionContext(), cmd, baseDn, null, false, false);
                embeddedMmds = LDAPUtils.getAllMemberMetaData(cmd);
                for (Attributes embeddedAttrs : entries.values())
                {
                    ObjectProvider embeddedSM = om.newObjectProviderForMember(mmd, new Type(om.getClassLoaderResolver().classForName(cmd.getFullClassName())));
                    embeddedSM = getEmbeddedStateManager(embeddedSM.getObject());
                    Object value = fetchMerge(embeddedSM, embeddedAttrs, embeddedMmds, embeddedMetaData);
                    if (value != null)
                    {
                        coll.add(value);
                    }
                }
            }
        }
        
        return coll;
    }

    private Object fetchMerge(ObjectProvider embeddedSM, Attributes embeddedAttrs, List<AbstractMemberMetaData> embeddedMmds,
            EmbeddedMetaData embeddedMetaData)
    {
        if (embeddedAttrs != null && embeddedMmds != null)
        {
            AbstractClassMetaData embeddedCmd = embeddedSM.getClassMetaData();
            String nullIndicatorColumn = embeddedMetaData.getNullIndicatorColumn();
            String nullIndicatorValue = embeddedMetaData.getNullIndicatorValue();
            // PK and owner first
            for (AbstractMemberMetaData embeddedMmd : embeddedMmds)
            {
                String fieldName = embeddedMmd.getName();
                int i = embeddedCmd.getAbsolutePositionOfMember(fieldName);

                if (fieldName.equals(embeddedMetaData.getOwnerMember()))
                {
                    Object embeddedValue = sm.getObject();
                    embeddedSM.replaceField(i, embeddedValue);
                }
                else if (embeddedMmd.isPrimaryKey())
                {
                    AbstractMappingStrategy ms = AbstractMappingStrategy.findMappingStrategy(storeMgr, embeddedSM, embeddedMmd, embeddedAttrs);
                    Object embeddedValue = ms.fetch();
                    embeddedSM.replaceField(i, embeddedValue);
                }
            }
            // Then other fields
            for (AbstractMemberMetaData embeddedMmd : embeddedMmds)
            {
                String fieldName = embeddedMmd.getName();
                String attributeName = LDAPUtils.getAttributeNameForField(embeddedMmd);
                int i = embeddedCmd.getAbsolutePositionOfMember(fieldName);

                if (!fieldName.equals(embeddedMetaData.getOwnerMember()) && !embeddedMmd.isPrimaryKey())
                {
                    AbstractMappingStrategy ms = AbstractMappingStrategy.findMappingStrategy(storeMgr, embeddedSM, embeddedMmd, embeddedAttrs);
                    Object embeddedValue = ms.fetch();

                    if (nullIndicatorColumn != null && attributeName.equals(nullIndicatorColumn))
                    {
                        if (embeddedValue == null)
                        {
                            return null;
                        }
                        if (nullIndicatorValue != null && embeddedValue.equals(nullIndicatorValue))
                        {
                            return null;
                        }
                    }

                    embeddedSM.replaceField(i, embeddedValue);
                }
            }
            return embeddedSM.getObject();
        }
        else
        {
            return null;
        }
    }

    private Collection<Object> getCollectionInstance(Class<?> collectionType)
    {
        Collection<Object> coll;
        try
        {
            coll = (Collection<Object>) collectionType.newInstance();
        }
        catch (InstantiationException e)
        {
            throw new NucleusException("Error in trying to create object of type " + collectionType.getName(), e);
        }
        catch (IllegalAccessException e)
        {
            throw new NucleusException("Error in trying to create object of type " + collectionType.getName(), e);
        }
        return coll;
    }

    private ObjectProvider getEmbeddedStateManager(Object value)
    {
        return om.findObjectProviderForEmbedded(value, sm, mmd);
    }

    @Override
    public void insert(Object value)
    {
        if (value == null)
        {
            return;
        }

        switch (mmd.getRelationType(clr))
        {
            case Relation.ONE_TO_ONE_UNI :
            case Relation.ONE_TO_ONE_BI :
            {
                Set<String> objectClasses = LDAPUtils.getObjectClassesForClass(effectiveClassMetaData);
                if (objectClasses.isEmpty())
                {
                    // embedded into the current entry
                    insertEmbedded(value);
                }
                else
                {
                    // embedded as child-entry
                    EmbeddedMetaData embeddedMetaData = mmd.getEmbeddedMetaData();
                    insertAsChild(value, embeddedMetaData);
                }
                break;
            }
            case Relation.ONE_TO_MANY_UNI :
            case Relation.ONE_TO_MANY_BI :
            {
                if (mmd.hasCollection())
                {
                    // 1-N (collection) relation
                    EmbeddedMetaData embeddedMetaData = mmd.getElementMetaData().getEmbeddedMetaData();
                    Collection c = (Collection) value;
                    for (Object pc : c)
                    {
                        insertAsChild(pc, embeddedMetaData);
                    }
                }
                break;
            }
            default :
            {
                // TODO Localise this
                throw new NucleusException(
                        "Field " + mmd.getFullFieldName() + " cannot be persisted because type=" + mmd.getTypeName() + " with relation type " + mmd
                                .getRelationType(clr) + " is not supported for this datastore");
            }
        }
    }

    private void insertEmbedded(Object value)
    {
        ObjectProvider embeddedSM = getEmbeddedStateManager(value);
        EmbeddedMetaData embeddedMetaData = mmd.getEmbeddedMetaData();
        List<AbstractMemberMetaData> embeddedMmds = new ArrayList<AbstractMemberMetaData>(Arrays.asList(embeddedMetaData
                .getMemberMetaData()));
        insertMerge(embeddedSM, attributes, embeddedMmds, embeddedMetaData);
    }

    private void insertAsChild(Object value, EmbeddedMetaData embeddedMetaData)
    {
        // merge fields
        ObjectProvider embeddedSM = getEmbeddedStateManager(value);
        effectiveClassMetaData = embeddedSM.getClassMetaData();
        List<AbstractMemberMetaData> embeddedMmds = LDAPUtils.getAllMemberMetaData(effectiveClassMetaData);
        BasicAttributes embeddedAttributes = new BasicAttributes();

        // split embedded fields and non-embedded fields
        List<AbstractMemberMetaData> nonEmbeddedMmds = new ArrayList<AbstractMemberMetaData>();
        for (Iterator<AbstractMemberMetaData> it = embeddedMmds.iterator(); it.hasNext();)
        {
            AbstractMemberMetaData mmd = it.next();
            if (!LDAPUtils.isEmbeddedField(mmd))
            {
                nonEmbeddedMmds.add(mmd);
                it.remove();
            }
        }

        // 1st: non-embedded members
        insertMerge(embeddedSM, embeddedAttributes, nonEmbeddedMmds, embeddedMetaData);

        // add object classes
        BasicAttribute objectClass = new BasicAttribute("objectClass");
        Set<String> embeddedObjectClasses = LDAPUtils.getObjectClassesForClass(effectiveClassMetaData);
        for (String oc : embeddedObjectClasses)
        {
            objectClass.add(oc);
        }
        embeddedAttributes.put(objectClass);

        // create entry
        LdapName dn = LDAPUtils.getDistinguishedNameForObject(storeMgr, embeddedSM);
        LDAPUtils.insert(storeMgr, dn, embeddedAttributes, om);

        // 2nd: embedded members
        insertMerge(embeddedSM, embeddedAttributes, embeddedMmds, embeddedMetaData);
    }

    private void insertMerge(ObjectProvider embeddedSM, Attributes embeddedAttributes, List<AbstractMemberMetaData> embeddedMmds,
            EmbeddedMetaData embeddedMetaData)
    {
        AbstractClassMetaData embeddedCmd = embeddedSM.getClassMetaData();

        // PK and owner first
        for (AbstractMemberMetaData embeddedMmd : embeddedMmds)
        {
            String fieldName = embeddedMmd.getName();
            int i = embeddedCmd.getAbsolutePositionOfMember(fieldName);

            if (fieldName.equals(embeddedMetaData.getOwnerMember()))
            {
                Object embeddedValue = sm.getObject();
                embeddedSM.replaceField(i, embeddedValue);
            }
            else if (embeddedMmd.isPrimaryKey())
            {
                AbstractMappingStrategy ms = AbstractMappingStrategy.findMappingStrategy(storeMgr, embeddedSM, embeddedMmd, embeddedAttributes);
                Object embeddedValue = embeddedSM.provideField(i);
                ms.insert(embeddedValue);
            }
        }

        for (AbstractMemberMetaData embeddedMmd : embeddedMmds)
        {
            String fieldName = embeddedMmd.getName();
            int i = embeddedCmd.getAbsolutePositionOfMember(fieldName);

            if (!fieldName.equals(embeddedMetaData.getOwnerMember()) && !embeddedMmd.isPrimaryKey())
            {
                AbstractMappingStrategy ms = AbstractMappingStrategy.findMappingStrategy(storeMgr, embeddedSM, embeddedMmd, embeddedAttributes);
                Object embeddedValue = embeddedSM.provideField(i);
                ms.insert(embeddedValue);
            }
        }
    }

    @Override
    public void update(Object value)
    {
        // System.out.println("Update field " + mmd.getName() + " of " + sm.getObject());
        switch (mmd.getRelationType(clr))
        {
            case Relation.ONE_TO_ONE_UNI :
            case Relation.ONE_TO_ONE_BI :
            {
                Set<String> objectClasses = LDAPUtils.getObjectClassesForClass(effectiveClassMetaData);
                if (objectClasses.isEmpty())
                {
                    // embedded into the current entry
                    updateEmbedded(value);
                }
                else
                {
                    // embedded as child-entry
                    // fetch old value, need to replace it afterwards
                    Object oldValue = fetchFromChild();
                    sm.replaceField(fieldNumber, value);
                    EmbeddedMetaData embeddedMetaData = mmd.getEmbeddedMetaData();
                    if (value == null)
                    {
                        if (oldValue != null)
                        {
                            // delete
                            deleteAsChild(oldValue, embeddedMetaData);
                        }
                    }
                    else
                    {
                        if (oldValue == null)
                        {
                            // insert
                            insertAsChild(value, embeddedMetaData);
                        }
                        else
                        {
                            ObjectProvider valueSM = getEmbeddedStateManager(value);
                            ObjectProvider oldValueSM = getEmbeddedStateManager(oldValue);
                            LdapName valueDn = LDAPUtils.getDistinguishedNameForObject(storeMgr, valueSM);
                            LdapName oldValueDn = LDAPUtils.getDistinguishedNameForObject(storeMgr, oldValueSM);
                            if (!valueDn.equals(oldValueDn))
                            {
                                // delete old, insert new
                                deleteAsChild(oldValue, embeddedMetaData);
                                insertAsChild(value, embeddedMetaData);
                            }
                            else
                            {
                                // update
                                updateAsChild(value, embeddedMetaData);
                            }
                        }
                    }
                }
                break;
            }
            case Relation.ONE_TO_MANY_UNI :
            case Relation.ONE_TO_MANY_BI :
            {
                if (mmd.hasCollection())
                {
                    // 1-N (collection) relation
                    Collection<Object> coll = (Collection<Object>) value;
                    LinkedHashMap<LdapName, Object> newMap = new LinkedHashMap<LdapName, Object>();
                    for (Object newObject : coll)
                    {
                        ObjectProvider newObjectSM = getEmbeddedStateManager(newObject);
                        LdapName newObjectDn = LDAPUtils.getDistinguishedNameForObject(storeMgr, newObjectSM);
                        newMap.put(newObjectDn, newObject);
                    }

                    EmbeddedMetaData embeddedMetaData = mmd.getElementMetaData().getEmbeddedMetaData();
                    Class oldCollInstanceType = org.datanucleus.store.types.sco.SCOUtils.getContainerInstanceType(mmd.getType(),
                        mmd.getOrderMetaData() != null);
                    Collection<Object> oldColl = fetchFromChildren(oldCollInstanceType, mmd, embeddedMetaData);
                    LinkedHashMap<LdapName, Object> oldMap = new LinkedHashMap<LdapName, Object>();
                    for (Object oldObject : oldColl)
                    {
                        ObjectProvider oldObjectSM = getEmbeddedStateManager(oldObject);
                        LdapName oldObjectDn = LDAPUtils.getDistinguishedNameForObject(storeMgr, oldObjectSM);
                        oldMap.put(oldObjectDn, oldObject);
                    }

                    Collection<Object> toInsert = null;
                    Collection<Object> toUpdate = null;
                    Collection<Object> toDelete = null;
                    Class instanceType = mmd.getType();
                    if (List.class.isAssignableFrom(instanceType))
                    {
                        toInsert = new ArrayList<Object>();
                        toUpdate = new ArrayList<Object>();
                        toDelete = new ArrayList<Object>();
                    }
                    else
                    {
                        toInsert = new HashSet<Object>();
                        toUpdate = new HashSet<Object>();
                        toDelete = new HashSet<Object>();
                    }

                    Iterator<LdapName> it = oldMap.keySet().iterator();
                    while (it.hasNext())
                    {
                        LdapName dn = it.next();
                        if (newMap.containsKey(dn))
                        {
                            toUpdate.add(newMap.get(dn));
                            newMap.remove(dn);
                            it.remove();
                        }
                    }
                    toInsert.addAll(newMap.values());
                    toDelete.addAll(oldMap.values());

                    for (Object pc : toInsert)
                    {
                        insertAsChild(pc, embeddedMetaData);
                    }
                    for (Object pc : toUpdate)
                    {
                        updateAsChild(pc, embeddedMetaData);
                    }
                    for (Object pc : toDelete)
                    {
                        deleteAsChild(pc, embeddedMetaData);
                    }
                }
                break;
            }
            default :
            {
                // TODO Localise this
                throw new NucleusException(
                        "Field " + mmd.getFullFieldName() + " cannot be persisted because type=" + mmd.getTypeName() + " with relation type " + mmd
                                .getRelationType(clr) + " is not supported for this datastore");
            }
        }
    }

    private void updateAsChild(Object value, EmbeddedMetaData embeddedMetaData)
    {
        // update
        ObjectProvider embeddedSM = om.findObjectProvider(value);
        boolean insert = false;
        if (embeddedSM == null)
        {
            embeddedSM = getEmbeddedStateManager(value);
            insert = true;
        }

        effectiveClassMetaData = embeddedSM.getClassMetaData();

        // need to update dirty fields *and* embedded fields
        int[] dirtyFieldNumbers = embeddedSM.getDirtyFieldNumbers();
        List<AbstractMemberMetaData> dirtyAndEmbeddedMmds = LDAPUtils.getMemberMetaData(dirtyFieldNumbers, effectiveClassMetaData);
        List<AbstractMemberMetaData> allMmds = LDAPUtils.getAllMemberMetaData(effectiveClassMetaData);
        for (AbstractMemberMetaData mmd : allMmds)
        {
            if (LDAPUtils.isEmbeddedField(mmd) && !dirtyAndEmbeddedMmds.contains(mmd))
            {
                dirtyAndEmbeddedMmds.add(mmd);
            }
        }
        if (dirtyAndEmbeddedMmds.isEmpty())
        {
            return;
        }

        // merge fields into the BasicAttributes data structure,
        // this also updates nested embedded objects recursively
        BasicAttributes embeddedAttributes = new BasicAttributes();
        updateMerge(embeddedSM, embeddedAttributes, dirtyAndEmbeddedMmds, embeddedMetaData, insert);

        // schedule update of fields
        if (embeddedAttributes.size() > 0)
        {
            // get DN
            LdapName dn = LDAPUtils.getDistinguishedNameForObject(storeMgr, embeddedSM);
            LDAPUtils.update(storeMgr, dn, embeddedAttributes, om);
        }
    }

    private void deleteAsChild(Object oldValue, EmbeddedMetaData embeddedMetaData)
    {
        ObjectProvider embeddedSM = getEmbeddedStateManager(oldValue);
        LdapName dn = LDAPUtils.getDistinguishedNameForObject(storeMgr, embeddedSM);
        LDAPUtils.deleteRecursive(storeMgr, dn, om);
    }

    private void updateEmbedded(Object value)
    {
        if (value == null)
        {
            // create an instance with empty fields, this will null-out all embedded fields
            ObjectProvider embeddedSM = om.newObjectProviderForMember(mmd, effectiveClassMetaData);
            int[] allMemberPositions = embeddedSM.getClassMetaData().getAllMemberPositions();
            for (int i : allMemberPositions)
            {
                embeddedSM.makeDirty(i);
            }
            //get the dummy value
            value = embeddedSM.getObject();
        }

        ObjectProvider embeddedSM = om.findObjectProvider(value);
        boolean insert = false;
        if (embeddedSM == null)
        {
            embeddedSM = getEmbeddedStateManager(value);
            insert = true;
        }

        EmbeddedMetaData embeddedMetaData = mmd.getEmbeddedMetaData();
        List<AbstractMemberMetaData> embeddedMmds = new ArrayList<AbstractMemberMetaData>(Arrays.asList(embeddedMetaData
                .getMemberMetaData()));
        updateMerge(embeddedSM, attributes, embeddedMmds, embeddedMetaData, insert);
    }

    private void updateMerge(ObjectProvider embeddedSM, Attributes embeddedAttributes, List<AbstractMemberMetaData> embeddedMmds,
            EmbeddedMetaData embeddedMetaData, boolean insert)
    {
        AbstractClassMetaData embeddedCmd = embeddedSM.getClassMetaData();

        // PK and owner first
        for (AbstractMemberMetaData embeddedMmd : embeddedMmds)
        {
            String fieldName = embeddedMmd.getName();
            int i = embeddedCmd.getAbsolutePositionOfMember(fieldName);

            if (fieldName.equals(embeddedMetaData.getOwnerMember()))
            {
                Object embeddedValue = sm.getObject();
                embeddedSM.replaceField(i, embeddedValue);
            }
            else if (embeddedMmd.isPrimaryKey())
            {
                AbstractMappingStrategy ms = AbstractMappingStrategy.findMappingStrategy(storeMgr, embeddedSM, embeddedMmd, embeddedAttributes);
                Object embeddedValue = embeddedSM.provideField(i);
                ms.update(embeddedValue);
            }
        }

        String[] dirtyFieldNames = embeddedSM.getDirtyFieldNames();
        List<String> dirtyFieldNameList = dirtyFieldNames != null ? Arrays.asList(dirtyFieldNames) : new ArrayList<String>();

        for (AbstractMemberMetaData embeddedMmd : embeddedMmds)
        {
            String fieldName = embeddedMmd.getName();
            if (!fieldName.equals(embeddedMetaData.getOwnerMember()) && !embeddedMmd.isPrimaryKey())
            {
                int i = embeddedCmd.getAbsolutePositionOfMember(fieldName);
                AbstractMappingStrategy ms = AbstractMappingStrategy.findMappingStrategy(storeMgr, embeddedSM, embeddedMmd, embeddedAttributes);
                Object embeddedValue = embeddedSM.provideField(i);
                if (insert || LDAPUtils.isEmbeddedField(embeddedMmd) || dirtyFieldNameList.contains(fieldName))
                {
                    ms.update(embeddedValue);
                }
            }
        }
    }

    @Override
    public List<String> getAttributeNames()
    {
        // no results for interface or abstract
        Class c = om.getClassLoaderResolver().classForName(effectiveClassMetaData.getFullClassName());
        if (c.isInterface() || Modifier.isAbstract(c.getModifiers()))
        {
            return Collections.EMPTY_LIST;
        }

        List<String> names = new ArrayList<String>();

        EmbeddedMetaData embeddedMetaData = null;
        switch (mmd.getRelationType(clr))
        {
            case Relation.ONE_TO_ONE_UNI :
            case Relation.ONE_TO_ONE_BI :
                embeddedMetaData = mmd.getEmbeddedMetaData();
                break;

            case Relation.ONE_TO_MANY_UNI :
            case Relation.ONE_TO_MANY_BI :
                embeddedMetaData = mmd.getElementMetaData().getEmbeddedMetaData();
        }

        
        if (embeddedMetaData != null)
        {
            List<AbstractMemberMetaData> embeddedMmds = new ArrayList<AbstractMemberMetaData>(Arrays.asList(embeddedMetaData
                    .getMemberMetaData()));
            ObjectProvider embeddedSM = om.newObjectProviderForMember(mmd, effectiveClassMetaData);
            for (AbstractMemberMetaData embeddedMmd : embeddedMmds)
            {
                AbstractMappingStrategy ms = AbstractMappingStrategy.findMappingStrategy(storeMgr, embeddedSM, embeddedMmd, new BasicAttributes());
                if (ms != null)
                {
                    List<String> embeddedAttributeNames = ms.getAttributeNames();
                    names.addAll(embeddedAttributeNames);
                }
            }
        }

        return names;
    }
}
